class Catalog
  @@catalog = {}

  def get_chain(name, nested = false)
    chain = @catalog[name]
    if chain.nil?
      raise RuntimeError, name + ' not found'
    end
    Chain.new(chain, nested)
  end

  def self.load_catalog(catfile)
    catalog = nil
    File.open(catfile, 'r') do |f|
      eval(f.read)
    end
    if catalog.nil? || catalog[:name].nil?
      raise RuntimeError, catfile + ' is not a valid catalog'
    else
      @@catalog[catalog[:name]] = catalog
    end
    Catalog.new(catalog)
  end

  def self.get_catalog(name)
    Catalog.new(@@catalog[name])
  end

 private
  def self.new(c)
    super(c)
  end
  def initialize(cat)
    @catalog = cat
  end

end

class Chain

  class Command
    def initialize(c)
      @obj = c[:class].new
      @id = c[:id]
    end
    def execute(ctx)
      @obj.execute(ctx)
    end
    def post_execute(cx, e)
      if @obj.respond_to?(:post_execute)
        @obj.post_execute(cx, e)
      end
    end
  end

  def initialize(cmd, nested)
    @nested = nested
    @commands = []
    cmd.each do |c|
      if Class === c[:class]
        @commands << Command.new(c)
      else
        jump = nil
        if !c[:catalog].nil?
          cat = Catalog.get_catalog(c[:catalog])
          if !cat.nil? && !c[:name].nil?
            jump = cat.get_chain(c[:name], true)
          end
        end
        if jump.nil? 
          if !c[:ooptional]
            raise RuntimeError, c.inspect + ' is not valid chain'
          end
        else
          @commands << jump
        end
      end
    end
    @bt = []
  end

  def execute(cx)
    e = nil
    result = nil
    @commands.each do |c|
      @bt << c
      begin
        result = c.execute(cx)
        break if result == true
      rescue StandardError => e
        if @nested
          raise e
        end
        break
      end
    end
    post_execute(cx, e) unless @nested
    result
  end

  def post_execute(cx, e)
    @bt.reverse_each do |x|
      begin
        if x.respond_to?(:post_execute)
          x.post_execute(cx, e)
        end
      rescue
        # ignore exception
      end
    end
  end
end
